package com.fossgalaxy.games.tbs.order;

import com.fossgalaxy.games.tbs.GameState;
import com.fossgalaxy.games.tbs.entity.Entity;
import com.fossgalaxy.games.tbs.metrics.GameMetrics;
import com.fossgalaxy.games.tbs.rules.Rule;
import com.fossgalaxy.games.tbs.ui.GameAction;

import java.util.*;
import java.util.stream.Collectors;

/**
 * Created by webpigeon on 22/01/18.
 */
public class OrderProcessor {

    private int turnNumber;
    private int currPlayer;
    private int numPlayers;
    private Integer winner;
    private Set<Integer> losers;

    private Set<Entity> entitiesNeedingOrders;
    private GameState state;

    private List<Rule> rules;

    private GameMetrics metrics;
    
    public OrderProcessor(GameState state, int numPlayers) {
        this(state, numPlayers, new GameMetrics());
    }

    public OrderProcessor(GameState state, int numPlayers, GameMetrics metrics) {
        this.metrics = metrics;

        this.entitiesNeedingOrders = new HashSet<>();
        this.rules = new ArrayList<>();
        this.losers = new HashSet<>();
        this.state = state;
        this.turnNumber = 0;
        this.currPlayer = 0;
        this.numPlayers = numPlayers;
    }

    public void setupTurn() {
        entitiesNeedingOrders.addAll(state.getOwnedEntities(currPlayer));

        //any environmental/global stuff "at start of turn"
        if (metrics != null) {
            metrics.onTurnSetup(this, state, currPlayer);
        }
    }

    public void doOrder(UUID entityID, Order order) {
        Entity entity = state.getEntityByID(entityID);

        //ignore illegal unit orders...
        if (entity == null) {
            return;
        }

        if (!entitiesNeedingOrders.contains(entity)) {
            throw new IllegalArgumentException("not that entity's turn!");
        }

        if (entity.getHealth() <= 0) {
            System.err.println("entity is dead, won't move");
            // entity was dead when order issued
            // could happen accidentally if the unit died during it's turn or the action didn't reap properly.
            return;
        }

        entitiesNeedingOrders.remove(entity);
        order.doOrder(entity, state);

        if (metrics != null) {
            metrics.onOrder(entityID, order, state);
        }
    }

    public Map<UUID, Order> doAutomaticActions() {
    	Map<UUID, Order> orders = new LinkedHashMap<>();
    	
    	Iterator<Entity> entityItr = entitiesNeedingOrders.iterator();
        while (entityItr.hasNext()) {
        	Entity entity = entityItr.next();
        	
            for (GameAction act : entity.getType().getAvailableActions()) {

            	//FIXME there is no check here for legal actions - I could add it but it'll probably break some of the games
                if (act.canAutomate()) {
                	
                    Order order = act.generateOrder(entity.getPos(), state);
                    
                	orders.put(entity.getID(), order);

                	if (metrics != null) {
                        metrics.onAutomaticOrder(entity.getID(), order, state);
                    }

                    order.doOrder(entity, state);
                	
                    break;
                }

            }

        }
        
        return orders;
    }

    public void skipAllRemaining() {
        if (metrics != null) {
            metrics.logIdleEntities(entitiesNeedingOrders);
        }
        entitiesNeedingOrders.clear();
    }

    /**
     * Tell the game that it's the next player's turn.
     * <p>
     * Any unit not issued with an order will be skipped.
     */
    public void finishTurn() {
        if (!isTurnOver()) {
            skipAllRemaining();
        }

        List<Entity> reapList = state.getEntities().stream()
                .filter(x -> x.getHealth() <= 0)
                .collect(Collectors.toList());
        reapList.forEach(x -> state.removeEntity(x));
        
        //any environmental/global stuff "at end of turn".
        for (Rule rule : rules) {
            Integer ruleWinner = rule.getWinner(state);
            if (ruleWinner != Rule.NO_WINNER) {
                winner = ruleWinner;

                if (metrics != null) {
                    metrics.logWinCondition(rule);
                }
                break;
            }
            List<Integer> ruleLosers = rule.getLosers(state);
            if (!ruleLosers.isEmpty()) {
                System.out.println("Losers: " + ruleLosers);
                losers.addAll(ruleLosers);
                if(state.getNumberOfPlayers() - 1 == ruleLosers.size()){
                    // Find player that isn't in ruleLosers
                    for(int i = 0; i < state.getNumberOfPlayers(); i++){
                        if(losers.contains(i)) continue;
                        winner = i;
                        return;
                    }
                }else if(state.getNumberOfPlayers() == ruleLosers.size()){
                    winner = currPlayer;
                    return;
                }
            }
        }

        if (metrics != null) {
            metrics.onTurnEnd(state, winner, losers);
        }

        //update the turn counter skipping losers
        state.tick();
        turnNumber++;
        
        while (true) {
            currPlayer = (currPlayer + 1) % numPlayers;
            if (!losers.contains(currPlayer)) break;
        }
    }

    public Integer getWinner() {
        return winner;
    }

    public Collection<Entity> getEntitiesNeedingOrders() {
        return Collections.unmodifiableCollection(entitiesNeedingOrders);
    }

    public boolean isTurnOver() {
        return entitiesNeedingOrders.isEmpty();
    }

    public void doOrderBulk(Map<UUID, Order> orders) {

        for (Map.Entry<UUID, Order> orderEntry : orders.entrySet()) {
            doOrder(orderEntry.getKey(), orderEntry.getValue());
        }
        
        doAutomaticActions();
        skipAllRemaining();
    }

    public int getTurnNumber() {
        return turnNumber;
    }

    public int getCurrentPlayer() {
        return currPlayer;
    }

    public void addRule(Rule rule) {
        rules.add(rule);
    }
    
    public GameMetrics getMetrics() {
    	return metrics;
    }
}
